# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
"""
This module groups functions directly related to the design of an interaction
weighted graph based on the search of molecules of interest.

Entry point: :meth:`~cadbiom_cmd.interaction_graph.json_2_interaction_graph`.
"""
from __future__ import unicode_literals
from __future__ import print_function

# Standard imports
import time
import json
import os
import glob
import itertools as it
from collections import defaultdict, Counter
import networkx as nx

# Library imports
import cadbiom.commons as cm

LOGGER = cm.logger()


## Handle *.json files #########################################################

def get_solutions_and_related_places_from_file(file_path):
    """Get frontier places and other places involved in transitions.


    :param file_path: Path of a JSON file; this file is generated by
        :meth:`~cadbiom_cmd.tools.solutions.convert_solutions_to_json`.

        :A solution is composed of steps with events, composed of transitions:

            .. code-block:: python

                [{
                    "solution": "Ax Bx",
                    "steps": [
                        [
                            {
                                "event": "_h_2",
                                "transitions": [{
                                    "ext": "n3",
                                    "ori": "Bx"
                                }]
                            },
                        ],
                    ]
                }]
    :param file_path: <str>
    :return: A generator of tuples with tuple of frontier places as keys
        and set of places involved in transitions as values.

        .. code-block:: python

            (("Ax", "Bx"), {"n3", "Bx"})
    :rtype: <generator <tuple <tuple>, <set>>>
    """

    decomp_solutions = json.load(open(file_path))

    # Get all places for each solution of the current problem
    for solution in decomp_solutions:
#        places = set()
#        # A solution is composed of steps
#        for step in solution['steps']:
#            # A step is composed of events
#            for event in step:
#                # An event is composed of 2 places: ori -> ext
#                for transition in event['transitions']:
#                    places.add(transition['ext'])
#                    places.add(transition['ori'])
#        yield (tuple(solution['solution'].split(' ')), places)

        yield (
            tuple(solution['solution'].split(' ')),
            set(it.chain(*((transition['ext'], transition['ori'])
                           for step in solution['steps']
                           for event in step
                           for transition in event['transitions']))))


def get_solutions_and_related_places(path):
    """Read decompiled solutions files (\*.json\* files)

    This functions tests if the given path is a directory or a file.

    :param path: Filepath/directory of a decompiled JSON file.
    :type path: <str>
    :return: A generator of tuples with tuple of frontier places as keys
        and set of places involved in transitions as values.

        .. code-block:: python

            (("Ax", "Bx"), {"n3", "Bx"})
    :rtype: <generator <tuple <tuple>, <set>>>
    """

    # Check valid input file/directory
    assert os.path.isfile(path) or os.path.isdir(path)

    LOGGER.info("Opening files...")

    # Get tuples of frontier places and set of places involved in transitions
    # for each file
    if os.path.isfile(path):
        # the given path is a file
        for sol_places in get_solutions_and_related_places_from_file(path):
            yield sol_places

    elif os.path.isdir(path):
        # The given path is a directory
        path = path if path[-1] == '/' \
            else path + '/'

        file_number = 0
        for file_number, file_path in \
            enumerate(glob.glob(path + '*_decomp.json'), 1):
            for sol_places in get_solutions_and_related_places_from_file(file_path):
                yield sol_places

        LOGGER.info("Files processed: %s", file_number)
        assert file_number != 0, "No *_decomp.json files found!"

## Build the graph ### #########################################################

def filter_trajectories(trajectories, molecules_of_interest):
    """Get solutions and count frontier places related to the given molecules
    of interest.

    :param trajectories: A generator of tuples with tuple of frontier places as
        keys and set of places involved in transitions as values.

        .. code-block:: python

            (("Ax", "Bx"), {"n3", "Bx"})
    :param molecules_of_interest: Iterable of molecules of interest.
    :type trajectories: <generator <tuple <tuple>, <set>>>
    :type molecules_of_interest: <tuple>
    :return: A tuple of all solutions related to the molecules of interest,
        and a dictionary of related frontier places and their occurences for
        each molecule of interest.

        .. code-block:: python

            # For molecules of interest "A" and "B"
            ((("frontier_1", "frontier_2", "frontier_3"),),
             {"A": {
                "frontier_1": 1,
                "frontier_2": 1,
              },
              "B": {
                "frontier_3": 1,
              },
             })
    :rtype: <tuple <tuple <tuple <str>>>, <dict <str>: <Counter <str>: <int>>>>
    """

    molecules_of_interest = set(molecules_of_interest)

    def entity_is_in_trajectory(sol, trajectory_places):
        """Search interactions between molecules of interest and entities
        involved in the given trajectory, and pre-build binary interactions.

        We assume that the molecules of interest **are** in the trajectories.
        We associate with each entity the trajectories containing them.

        :param trajectory_places: A trajectory: i.e a set of entities.
        :type trajectory_places: <set>
        :return: True if an entity of interest is found in the given trajectory.
        :rtype: <boolean>
        """
        entity_found = False

        # Search each entity of interest among the places in the trajectory
        g = (searched_molec for searched_molec in molecules_of_interest
             if searched_molec in trajectory_places)

        for searched_molec in g:
            # Associate the entity of interest with the frontier places and
            # count their occurences.
            # The count is kept in order to weight the edges of the graph.
            binary_interactions[searched_molec] += Counter(sol)#trajectory_places
            entity_found = True
        return entity_found


    binary_interactions = defaultdict(Counter)

    # Almost one entity of interest is found in the trajectory
    # => we keep the solution
    filtered_macs = tuple(sol for sol, trajectory_places in trajectories
                          if entity_is_in_trajectory(sol, trajectory_places))

    assert filtered_macs, "No molecule of interest was found."

    return filtered_macs, binary_interactions


def build_interactions(filtered_macs, binary_interactions):
    """Make binary interactions used by the graph as edges

    PS: genes and stimulis are frontier places.

    :param filtered_macs: All solutions related to the molecules of interest.

        .. code-block:: python

            (("frontier_1", "frontier_2", "frontier_3"),)
    :type filtered_macs: <tuple <tuple <str>>>
    :param binary_interactions: A dictionary of related frontier places.

        .. code-block:: python

            # For molecules of interest "A" and "B"
            {"A": {
               "frontier_1": 1,
               "frontier_2": 1,
             },
             "B": {
               "frontier_3": 1,
             },
            }
    :type binary_interactions: <dict <str>: <Counter <str>: <int>>>
    :return: Various Counters of binary interactions:

        * all_genes: All genes in all the solutions
        * all_stimuli: All stimulis in all the solutions
        * genes_interactions: Interactions between genes in the same solution
        * stimulis_interactions: Interactions between stimuli in the same solution
        * genes_stimuli_interactions: Interactions between genes and stimulis in
          the same solution
        * molecule_stimuli_interactions: Counter interactions between molecules
          of interest and frontier places that are not genes (stimuli)
          in trajectories (i.e.: ``(molecule, stimulus)``).

    :rtype: <set>, <set>, <Counter>, <Counter>, <Counter>, <Counter>
    """

    genes_stimuli_interactions = Counter()
    genes_interactions = Counter()
    stimulis_interactions = Counter()
    # All genes in frontier places
    all_genes = set()
    # All other entities in frontier places
    all_stimuli = set()

    for frontier_places in filtered_macs:

        # Extract genes from the frontier places
        genes = {frontier_place for frontier_place in frontier_places
                 if '_gene' in frontier_place}
        # Stimulis are the other places
        stimuli = set(frontier_places) - genes

        all_genes.update(genes)
        all_stimuli.update(stimuli)


        # For every solution composed of frontier places,
        # compute binary interactions between:
        # genes and stimulis
        genes_stimuli_interactions.update(it.product(genes, stimuli))
        # genes two by two
        genes_interactions.update(it.combinations(genes, 2))
        # stimulis two by two
        stimulis_interactions.update(it.combinations(stimuli, 2))

#    print("genes/stimulis interactions:", genes_stimuli_interactions)
#    print("genes interactions:", genes_interactions)
#    print("stimulis interactions:", stimulis_interactions)

    LOGGER.debug("build_interactions:: all genes:\n%s", all_genes)
    LOGGER.debug("build_interactions:: all stimuli:\n%s", all_stimuli)

    #######

    def make_molec_stimuli_interactions(molecule_of_interest, places_occurrences):
        """Make binary interactions between a molecule of interest and
        the stimuli in their solutions.

        We also add to each generated tuple, its number of occurences.

        We avoid all interactions that contain genes, and do not contain
        stimuli.

        .. warning:: Will not work if the molecule of interest is only in conditions
            and not in trajectories. In this case the molecule is a frontier
            place (an input of the model); so it seems useless to search this
            kind of entity.

        :param molecule_of_interest: A searched molecule (user input).
        :param places_occurrences: Counter with frontier placs as keys,
            and their occurences as values.
        :type molecule_of_interest: <str>
        :type places_occurrences: <Counter <str>: <int>>
        """

        frontier_places = set(places_occurrences.keys())

        # Remove interactions with a gene
        # => remove the link between the gene (frontier place) and the molecule of interest
        # => remove external stage of the graphe
        frontier_places -= all_genes

        # Remove interactions without stimuli (remove all entities that are not boundaries)
        # => Keep interaction between stimuli and molecule of interest
        # PS: frontier_places != all_stimuli because all_stimuli is a set
        # of all stimuli accross all solutions.
        # Here, frontier_places is a set of all solutions linked to the
        # given molecule_of_interest
        frontier_places &= all_stimuli


        # Get occurences for each binary interaction (molecule, frontier_place)
        return {molec_place: places_occurrences[molec_place[1]]
                for molec_place in it.product([molecule_of_interest], frontier_places)}


    g = (make_molec_stimuli_interactions(*data)
         for data in binary_interactions.iteritems())
    # warning: marchera pas si la molec of interest est dans les conditions
    # et non dans les places des trajectoires...
    # => Comme si la molecule n'était pas trouvée
    molecule_stimuli_interactions = Counter()
    [molecule_stimuli_interactions.update(f_bin) for f_bin in g]

    return all_genes, all_stimuli, genes_interactions, \
        stimulis_interactions, \
        genes_stimuli_interactions, \
        molecule_stimuli_interactions


def build_graph(output_dir, all_genes, all_stimuli, genes_interactions,
                stimulis_interactions,
                genes_stimuli_interactions,
                molecule_stimuli_interactions):
    """Make an interaction weighted graph based on the search of molecules of interest

    :Edges:

        * gene - gene: Two genes present simultaneously in a solution
        * stimulus - stimulus: Two stimuli present simultaneously in a solution
        * gene - stimulus: One gene and one stimulus present simultaneously in
          a solution (deprecated)
        * molecule of interest - stimulus: A molecule of interest in a trajectory
          related to a solution that contains a stimulus.

    :Legend of the edges:

        * gene - gene: red
        * stimulus - stimulus: blue (deprecated)
        * gene - stimulus: red
        * molecule of interest - stimulus: yellow

    :Legend of the nodes:

        * genes: red
        * stimuli: blue
        * molecules of interest: yellow

    :param output_dir: Output path.
    :param all_genes: All genes in all the solutions
    :param all_stimuli: All stimulis in all the solutions
    :param genes_interactions: Interactions between genes in the same solution
    :param stimulis_interactions: Interactions between stimuli in the same solution
    :param genes_stimuli_interactions: Interactions between genes and stimulis in
        the same solution
    :param molecule_stimuli_interactions: Counter interactions between molecules
        of interest and frontier places that are not genes (stimuli)
        in trajectories (i.e.: ``(molecule, stimulus)``).
    :type output_dir: <str>
    :type all_genes: <set>
    :type all_stimuli: <set>
    :type genes_interactions: <Counter>
    :type stimulis_interactions: <Counter>
    :type genes_stimuli_interactions: <Counter>
    :type molecule_stimuli_interactions: <Counter>
    """

    LOGGER.info("Building graph...")

    # Count all nodes
    nodes_count = Counter(it.chain(*genes_interactions.keys()))
    nodes_count.update(it.chain(*stimulis_interactions.keys()))
    nodes_count.update(it.chain(*genes_stimuli_interactions.keys()))
    nodes_count.update(it.chain(*molecule_stimuli_interactions.keys()))

    # Make Graph ###############################################################
    def colormap(node):
        """Get color assigned to the given node depending on its group"""
        if node in all_genes:
            return 'red'
        elif node in all_stimuli:
            return 'blue'
        # Molecule of interest
        return 'yellow'

    G = nx.DiGraph()

    # Add all nodes with their attributes: color & weight
    [G.add_node(node, weight=weight, color=colormap(node))
     for node, weight in nodes_count.iteritems()]

    def build_edge_data(edges):
        """Build edge tuple for networkx API"""
        return [(node1, node2, weight) for (node1, node2), weight in edges.iteritems()]

    # Add all edges
    # PS Syntax: G.add_weighted_edges_from([(0,1,3.0),(1,2,7.5)],color='red')
    # Interactions between genes in the same solution
    G.add_weighted_edges_from(
        build_edge_data(genes_interactions), color='red')
    # Interactions between stimuli in the same solution
    #G.add_weighted_edges_from(build_edge_data(stimulis_interactions), color='blue')
    # Interactions between genes and stimulis in the same solution
    G.add_weighted_edges_from(
        build_edge_data(genes_stimuli_interactions), color='red')
    # Interactions between molecules of interest and frontier places
    # that are not genes in trajectories
    G.add_weighted_edges_from(
        build_edge_data(molecule_stimuli_interactions), color='yellow')

    # Write graph
    nx.write_graphml(G, output_dir + "interaction_graph.graphml")


def json_2_interaction_graph(output_dir, molecules_of_interest, path):
    """Entry point for json_2_interaction_graph

    Read decompiled solutions files (\*.json\* files produced by the
    directive ``solutions_2_json``) and make a graph of the relationships
    between one or more molecules of interest, the genes and other
    frontier places/boundaries found among all the solutions.

    More information about the graph and its legend:
    :meth:`~cadbiom_cmd.interaction_graph.build_graph`.

    :param output_dir: Output path.
    :param molecules_of_interest: Iterable of molecules of interest.
    :param path: Filepath/directory of a JSON solution file.
    :type output_dir: <str>
    :type molecules_of_interest: <tuple>
    :type path: <str>
    """

    chrono_start = time.time()

    # get_solutions_and_related_places: trajectories
    # filter_trajectories: filtered_macs, binary_interactions
    # build_interactions: data to unpack for build_graph
    build_graph(
        output_dir,
        *build_interactions(
            *filter_trajectories(
                get_solutions_and_related_places(path),
                molecules_of_interest
            )
        )
    )

    LOGGER.info("Graph generated in %s", time.time() - chrono_start)





# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
"""
This module groups functions directly related to the management and the
extraction of data of a Cadbiom model.

Here we find high-level functions to manage the logical formulas of the events
and conditions defining the transitions; as well as useful functions to manage
the entities, like to obtain their metadata or the frontier places of the model.
"""
from __future__ import unicode_literals
from __future__ import print_function

# Standard imports
from collections import defaultdict
import re
import json
import itertools as it
from logging import DEBUG

# Library imports
from cadbiom.models.guard_transitions.translators.chart_xml \
    import MakeModelFromXmlFile
from cadbiom.models.biosignal.translators.gt_visitors import compile_event, compile_cond
from cadbiom.models.biosignal.sig_expr import *
from cadbiom.models.guard_transitions.analyser.ana_visitors import TableVisitor

import cadbiom.commons as cm

LOGGER = cm.logger()


class Reporter(object):
    """Error reporter.

    .. note:: Link the lexer to the model allows to avoid error in Reporter
        like:  "-> dec -> Undeclared event or state"
        In practice this is time consuming and useless for what we want to do.
        See parse_condition()
    """

    def __init__(self):
        self.error = False
        self.mess = ""

    def display(self, err):
        """Display the error in the logger"""
        self.error = True
        if "Undeclared event or state" not in err:
            LOGGER.error("\t" + self.mess + " -> " + err)


def get_transitions_from_model_file(model_file):
    """Get all transitions from a model file (bcx format).

    :param: bcx file.
    :type: <str>
    :return: Transitions (see get_transitions()).
    :rtype: <dict>
    """

    parser = MakeModelFromXmlFile(model_file)
    return get_transitions(parser)


def get_transitions(parser):
    """Get all transitions in the given parser.

    There are two methods to access the transitions of a model.

    :Example:

        .. code-block:: python

            >>> print(dir(parser))
            ['get_model', 'handler', 'model', 'parser']
            >>> # Direct access
            >>> events = list()
            >>> for transition in parser.get_model().transition_list:
            ...     events.append(transition.event)
            >>>
            >>> # Indirect access via a handler
            >>> events = list()
            >>> for transitions in parser.handler.top_pile.transitions:
            ...     # transitions is a list of CTransition objects
            ...     for transition in transitions:
            ...         events.append(transition.event)

    :param: Parser opened on a bcx file.
    :type: <MakeModelFromXmlFile>
    :return: A dictionnary of events as keys, and transitions as values.
        Since many transitions can define an event, values are lists.
        Each transition is a tuple with: origin node, final node, attributes
        like label and condition.
        ``{'h00': [('Ax', 'n1', {'label': 'h00[]'}),]``
    :rtype: <dict <list <tuple <str>, <str>, <dict <str>: <str>>>>
    """

    # NOTE: je devrais parler d'events au lieu de transitions...
    # voir si on peut retourner le parser pour faire tourner le static analysis ?
    # ou faire 1 fonction séparée qui parle plus du modèle lui meme que du graphe...
    # (ce que fait get_statistics d'ailleurs...)
    transitions = defaultdict(list)

    for trans in parser.get_model().transition_list:

        # Get the names of clocks
        # Some event have many clocks (like _h_2755) for the same
        # ori/ext entities, so we have to extract them and their respective
        # conditions
        if trans.event == '':
            # null event without clock => StartNodes
            # These nodes are used to resolve the problem of
            # Strongly Connected Components (inactivated cycles in the graph)
            # The nodes
            # Avoids having SigConstExpr as event type in parse_event()
            # I create a transition (SCC-__start__?),
            # and a node (__start__?) for this case.
            trans.event = 'SCC-' + trans.ori.name
            events = {trans.event: trans.condition}
        elif re.match('_h[\w]+', trans.event):
            # 1 event (with 1 clock)
            events = {trans.event: trans.condition}
        else:
            # Many events (with many clocks with condition(s))
            events = parse_event(trans.event)

        for event, condition in events.iteritems():
            # LOGGER.debug("NEW trans", event)

            # Handle multiple transitions for 1 event
            transitions[event].append(
                (
                    trans.ori.name, trans.ext.name,
                    {
                        'label': event, #+ '[' + trans.condition + ']',
                        'condition': condition,
                    }
                )
            )

    LOGGER.info("%s transitions loaded", len(transitions))
    # Return a dict instead of defaultdict to avoid later confusions
    #(masked errors) by searching a transition that was not in the model...

    assert transitions, "No transitions found in the model ! " \
        "Please check the names of events (_h_xxx)"

    # Forge return value
    return dict(transitions)


def get_frontier_places(transitions, all_places):
    """Return frontier places of a model (deducted from its transitions and
    from all places of the model).

    .. note:: why we use all_places from the model instead of
        (input_places - output_places) to get frontier places ?
        Because some nodes are only in conditions and not in transitions.
        If we don't do that, these nodes are missing when we compute
        valid paths from conditions.

    :param arg1: Model's transitions.
        {u'h00': [('Ax', 'n1', {u'label': u'h00[]'}),]
    :type arg1: <dict>
        keys: names of events
        values: list of transitions as tuples (with in/output, and label).
    :return: Set of frontier places.
    :rtype: <set>
    """

    # Get transitions in events
    g = tuple(trans for event in transitions.values() for trans in event)

    # Get input nodes & output nodes
#    input_places = {trans[0] for trans in g}
    output_places = {trans[1] for trans in g}

    # Get all places that are not in transitions in the "output" place
    return set(all_places) - output_places


################################################################################

def get_places_from_condition(condition):
    """Parse condition string and return all places, regardless of operators.

    .. note:: This function is only used to get all nodes in a condition when
        we know they are all inhibitors nodes.

    .. todo:: See the workaround in the code, without using very time consuming
        and badly coded functions.

    :param: Condition string.
    :type: <str>
    :return: Set of places.
    :rtype: <set>
    """

    # Valid but very time consuming like any other things in Cadbiom library
#    err = Reporter()
#    tvi = TableVisitor(err)
#    symb_tab = tvi.tab_symb
#    cond_sexpr = compile_cond(condition, symb_tab, err)
#    inhibitors_nodes = set()
#    possible_paths = rec(cond_sexpr, inhibitors_nodes)
#    return set(it.chain(*possible_paths))

    # Replace parentheses first to make spaces in the string
    # As operators are followed or preceded by parentheses, we can detect them
    # without false positives (operator string inside an entity name)
    replacement = ['(', ')', ' and ', ' or ', ' not ']

    for operator in replacement:
        condition = condition.replace(operator, ' ')

    # Must be exempt of unauthorized chars
    return {elem for elem in condition.split(' ')
            if elem != ''}


def parse_event(event):
    """Decompile logical formula in event's name.

    :param: Event string.
    :type: <event string>
    :return: A dict of events and their conditions.
    :rtype: <dict>
        keys: event's names; values: logical formula attached (condition)
    """

    def treeToExprDefaultsList(tree):
        if isinstance(tree, SigDefaultExpr):
            return treeToExprDefaultsList(tree.left_h) + \
                treeToExprDefaultsList(tree.right_h)

        # Here, some tree are from classes SigConstExpr or SigIdentExpr
        # Ex: for the clock "_h_5231":
        #    ... default (_h_5231)" => No condition for this event
        # Other examples:
        # _h_2018 _h_820 _h_4939 _h_5231 _h_3301 _h_4967 _h_2303 _h_3301
        return [tree]

    def filterSigExpression(expr):
        """
        .. note:: No SigConstExpr here => filtered in get_transitions()
            by checking null events (event="") in the model.
        """

        if isinstance(expr, SigWhenExpr):
            # right : SigSyncBinExpr (logical formula), BUT
            # sometimes SigConstExpr (just a True boolean) when clock is empty
            # Ex: "when ()"
            # So, we replace this boolean with an empty condition
            right = '' if isinstance(expr.right_h, SigConstExpr) \
                        else str(expr.right_h)

            return expr.left_h.name, right

        if isinstance(expr, SigIdentExpr):
            return expr.name, ''

        raise AssertionError("You should never have been there ! "
                             "Your expression type is not yet supported...")

#    def filterSigExpression(listOfExpr):
#        return [filterSigExpression(expr) for expr in listOfExpr]

    # Error Reporter
    err = Reporter()
    tvi = TableVisitor(err)
    symb_tab = tvi.tab_symb

    # Get tree object from event string
    event_sexpr = compile_event(event, symb_tab, True, err)[0]

    # Filter when events
    g = (filterSigExpression(expr) \
         for expr in treeToExprDefaultsList(event_sexpr))
    eventToCondStr = \
        {event_name: event_cond for event_name, event_cond in g}


    LOGGER.debug("Clocks from event parsing: %s", eventToCondStr)

    return eventToCondStr


def parse_condition(condition, all_nodes, inhibitors_nodes):
    """Return valid paths according the given logical formula and nodes;
    and set inhibitors_nodes

    .. note:: inhibitors_nodes is modified(set) by this function.

    :param condition: Condition string of a transition.
    :param all_nodes: Nodes involved in transitions + frontier places.
    :param inhibitors_nodes: Inactivated nodes in paths of conditions.
    :type condition: <str>
    :type inhibitors_nodes: <set>
    :type all_nodes: <set>
    :return: Set of paths. Each path is a tuple of nodes.
    :rtype: <set>
    """

    LOGGER.debug("CONDITION: %s", condition)
    # Error Reporter
    err = Reporter()
    tvi = TableVisitor(err)
    # Link the lexer to the model allows to avoid error in Reporter
    # like:  "-> dec -> Undeclared event or state"
    # In practice this is time consuming and useless for what we want to do
    #parser = MakeModelFromXmlFile(BIO_MOLDELS_DIR +
    # "Whole NCI-PID database translated into CADBIOM formalism(and).bcx")
    #parser.get_model().accept(tvi)
    symb_tab = tvi.tab_symb
    # Get tree object from condition string
    cond_sexpr = compile_cond(condition, symb_tab, err)
    # Get all possible paths from the condition
    possible_paths = rec(cond_sexpr, inhibitors_nodes)

    # Prune possible paths according to:
    # - Inhibitor nodes that must be removed because they will never
    # be in the graph.
    # - All nodes in transitions (ori -> ext) because we know all transitions
    # in the graph, so we know which entities can be choosen to validate a path.
    # - All frontier places, that are known entities that can be in conditions
    # (not only in ori/ext) of transitions.
    # So: authorized nodes = frontier_places + transition_nodes - inhibitor nodes
    valid_paths = {tuple(path) for path in possible_paths
                   if (set(path) - inhibitors_nodes).issubset(all_nodes)}

    # Debugging only
    if LOGGER.getEffectiveLevel() == DEBUG:
        LOGGER.debug("INHIBIT NODES: %s", inhibitors_nodes)
        LOGGER.debug("ALL NODES: %s", all_nodes)
        LOGGER.debug("POSSIBLE PATHS: %s", possible_paths)
        LOGGER.debug("VALID PATHS: %s", valid_paths)

        if len(valid_paths) > 1:
            LOGGER.debug("Multiple valid paths in the model for: %s:\n%s", condition, valid_paths)

        for path in possible_paths:
            pruned_places = set(path) - inhibitors_nodes
            isinsubset = pruned_places.issubset(all_nodes)
            LOGGER.debug("PRUNED PATH: %s, VALID: %s", pruned_places, isinsubset)

    assert valid_paths, "No valid path in the model for: " + str(condition)

    return valid_paths


    from cadbiom.models.guard_transitions.analyser.ana_visitors import SigExpIdCollectVisitor

    # condition expressions contains only node ident
    icv = SigExpIdCollectVisitor()
    lst1 = cond_sexpr.accept(icv)
    print(cond_sexpr)
    print(type(cond_sexpr))
    print(dir(cond_sexpr))
    print("LISTE", lst1)
#    <class 'cadbiom.models.biosignal.sig_expr.SigSyncBinExpr'>
#    'accept', 'get_signals', 'get_ultimate_signals', 'is_bot', 'is_clock',
# 'is_const', 'is_const_false', 'is_ident', 'left_h', 'operator', 'right_h', 'test_equal']

    print(cond_sexpr.get_signals())
#    print(cond_sexpr.get_ultimate_signals())
    print("LEFT", cond_sexpr.left_h)
    print("OPERATOR", cond_sexpr.operator)
    print("RIGHT", cond_sexpr.right_h)


#    ret = treeToTab(cond_sexpr)
#    [set([('((formule', True)])]
#    print("treeToTab", ret)
#    print(type(ret))
#    print(dir(ret))


def rec(tree, inhibitors_nodes):
    """Recursive function to decompile conditions

    :param tree:
        :Example of tree argument:

            .. code-block:: python

                tree = ('H', 'v', (
                    ('F', 'v', 'G'),
                    '^',
                    (
                        ('A', 'v', 'B'),
                        '^',
                        ('C', 'v', ('D', '^', 'E'))
                    )
                ))
    """

#    print("TREE", tree, type(tree), dir(tree))

    if isinstance(tree, str):  # terminal node
        path = [tree]
        solutions = [path]
        return solutions
    if isinstance(tree, SigNotExpr):
        # tree.operand: the entity, type: SigIdentExpr
        LOGGER.debug("NOT OPERAND: %s, %s", tree.operand, type(tree.operand))
        try:
            current_inhibitors = get_places_from_condition(tree.operand.__str__())
            inhibitors_nodes.update(current_inhibitors)
            LOGGER.debug("INHIBITORS found: %s", current_inhibitors)

            path = [tree.operand.name]
            solutions = [path]
            return solutions
        except AttributeError:
            tree = tree.operand

    if isinstance(tree, SigIdentExpr):
        path = [tree.name]
        solutions = [path]
        return solutions



    lch = tree.left_h
    op  = tree.operator
    rch = tree.right_h
#    print('OZCJSH:', lch, op, rch, sep='\t\t')
    lpaths = rec(lch, inhibitors_nodes)
    rpaths = rec(rch, inhibitors_nodes)
#    print('VFUENK:', lpaths, rpaths)
    if op == 'or':  # or
#        ret = [*lpaths, *rpaths]
        ret = list(it.chain(lpaths, rpaths))
#        print('RET:', ret)
        return ret
    else:  # and
        assert op == 'and'
#        print(list(it.product(lpaths, rpaths)))
#        raw_input('test')

        ret = list(l + r for l, r in it.product(lpaths, rpaths))
#        print('RET:', ret)
        return ret


################################################################################

def get_places_data(places, model):
    """Get a list of JSON data parsed from each given places in the model.

    .. note:: This function is used by low_model_infos().

    .. note:: v1 models return a dict with only 1 key: 'cadbiomName'

    .. note:: Start nodes (with a name like __start__x) are handled even
        with no JSON data.
        They are counted in the other_types and other_locations fields.

    :Example of JSON data that can be found in the model:

        .. code-block:: python

            {
                "uri": entity.uri,
                "entityType": entity.entityType,
                "names": list(entity.synonyms | set([entity.name])),
                "entityRef": entity.entityRef,
                "location": entity.location.name if entity.location else None,
                "modificationFeatures": dict(entity.modificationFeatures),
                "members": list(entity.members),
                "reactions": [reaction.uri for reaction in entity.reactions],
                "xrefs": entity.xrefs,
            }

    :param arg1: Iterable of name of places.
    :param arg2: Model from handler.
    :type arg1: <set>
    :type arg2: <MakeModelFromXmlFile>
    :return: List of data parsed from each give places.

        .. note:: Here is the list of field retrieved for v2 models:

            - cadbiomName
            - uri
            - entityType
            - entityRef
            - location
            - names
            - xrefs
    :rtype: <list <dict>>
    """

    if model.xml_namespace == 'http://cadbiom.genouest.org/v2/':
        # Fixed fields and default types
        json_note_fieldnames = {
            "uri": '',
            "entityType": '',
            "entityRef": '',
            "location": '',
            "names": list(),    # Default type is 'list' for names (not '')
            "xrefs": dict(),
        }
        # Init final dictionary
        data = list()
        for place_name in places:

            try:
                # Model type 2 => We use JSON data in each nodes
                # Get JSON data ('' if the field is not present)
                json_data = json.loads(model.node_dict[place_name].note)
            except ValueError as exc:
                # Handle start nodes (name: __start__x)
                if exc.message == 'No JSON object could be decoded':
                    json_data = dict()

            temp = \
                {fieldname: json_data.get(fieldname, default_data)
                 for fieldname, default_data in json_note_fieldnames.items()}

            # Patch: Handle null values that should be avoided in cadbiom_writer.build_json_data()
            temp["names"] = [name for name in temp["names"] if name]
            # Add the cadbiom name (name attribute of xml element
            temp['cadbiomName'] = place_name
            data.append(temp)

        return data


    # v1 model: return only the name of the place
    return [{'cadbiomName': place_name} for place_name in places]


def get_identifiers_mapping(model_file, external_identifiers):
    """Get Cadbiom names corresponding to the given external identifiers (xrefs)

    .. note:: This function works only on v2 formated models with JSON additional data

    :param model_file: Model file.
    :param external_identifiers: Set of external identifiers to be mapped.
    :type model_file: <str>
    :type external_identifiers: <set>
    :return: Mapping dictionary with external identifiers as keys
        and cadbiom names as values.
    :rtype: <dict <str>:<list>>
    """
    # Get the model
    parser = MakeModelFromXmlFile(model_file)
    model = parser.handler.model

    assert model.xml_namespace == 'http://cadbiom.genouest.org/v2/', \
        "Operation not supported: Only v2 models are supported."

    # Get all nodes
    places_data = get_places_data(parser.handler.node_dict.iterkeys(), model)

    # {'xrefs': {'bdd': [values],}, 'cadbiomName': '',}
    g = {place['cadbiomName']: frozenset(it.chain(*place['xrefs'].itervalues()))
         for place in places_data}

    # Mapping: external_identifiers as keys and Cadbiom names as values
    mapping = defaultdict(set)
    for place, identifiers in g.iteritems():

        common_identifiers = identifiers & external_identifiers
        if common_identifiers:
            [mapping[common_id].add(place) for common_id in common_identifiers]

    not_found_identifiers = external_identifiers - set(mapping.keys())
    if not_found_identifiers:
        LOGGER.info("Some identifiers were not found (%s/%s): %s",
                    len(not_found_identifiers),
                    len(external_identifiers),
                    not_found_identifiers
        )

    return mapping



# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
"""Entry point and argument parser for cadbiom_cmd package"""
from __future__ import print_function

# Standard imports
import argparse
import os
import sys
from functools import wraps

# Custom imports
import cadbiom.commons as cm

LOGGER = cm.logger()


def check_output_dir(function):
    """Decorator used process some arguments from argparse"""
    # Expose the docstring of the wrapped function instead of that of the decorator
    @wraps(function)
    def modified_func(*args, **kwargs):
        """Fix paths in arguments (add trailing '/')"""
        output_dir = args[0]['output']
        output_dir = output_dir if output_dir[-1] == '/' else output_dir + '/'

        # Compatibility for output and output_dir
        args[0]['output'] = output_dir
        args[0]['output_dir'] = output_dir

        return function(*args, **kwargs)

    return modified_func


@check_output_dir
def solutions_search(args):
    """Launch the search for Minimum Activation Conditions (MAC) for entities
    of interest.
    """

    # Module import
    import solution_search
    # Temporary fix: all_macs is always activated in the new API
    args["all_macs"] = True
    solution_search.solutions_search(args) # !


def solutions_sort(args):
    """Read a solution file or a directory containing MAC solutions files
    (*mac* files), and sort all frontier places/boundaries in alphabetical order.
    """

    # Module import
    import solution_sort
    solution_sort.solutions_sort(args['path'])


@check_output_dir
def solutions_2_graphs(args):
    """Create GraphML formated files containing a representation of the
    trajectories for each solution in complete MAC files (*mac_complete files).

    This is a function to visualize paths taken by the solver from the boundaries
    to the entities of interest.
    """

    # Module import
    import solution_sort
    solution_sort.solutions_2_graphs(
        args['output'],
        args['model_file'],
        args['path']
    )


@check_output_dir
def solutions_2_json(args):
    """Create a JSON formated file containing all data from complete MAC files
    (*mac_complete files). The file will contain frontier places/boundaries
    and decompiled steps with their respective events for each solution.

    This is a function to quickly search all transition attributes involved
    in a solution.
    """

    # Module import
    import solution_sort
    solution_sort.solutions_2_json(
        args['output'],
        args['model_file'],
        args['path'],
        conditions=not args['no_conditions'], # Reverse the param...
    )


@check_output_dir
def json_2_interaction_graph(args):
    """Make an interaction weighted graph based on the searched molecule of interest.

    Read decompiled solutions files (*.json* files produced by the
    directive 'solutions_2_json') and make a graph of the relationships
    between one or more molecules of interest, the genes and other
    frontier places/boundaries found among all the solutions.
    """

    # Module import
    import interaction_graph
    interaction_graph.json_2_interaction_graph(
        args['output'],
        args['molecules_of_interest'],
        args['path'],
    )


@check_output_dir
def solutions_2_common_graph(args):
    """Create a GraphML formated file containing a unique representation of **all**
    trajectories corresponding to all solutions in each complete MAC files
    (*mac_complete files).

    This is a function to visualize paths taken by the solver from the boundaries
    to the entities of interest.
    """

    # Module import
    import solution_sort
    solution_sort.solutions_2_common_graph(
        args['output'],
        args['model_file'],
        args['path']
    )


@check_output_dir
def solutions_2_occcurrences_matrix(args):
    """Create a matrix of occurrences counting entities in the solutions found in
    *mac.txt files in the given path.
    """

    # Module import
    import solution_sort
    solution_sort.solutions_2_occcurrences_matrix(
        args['output'],
        args['model_file'],
        args['path'],
        transposed=args['transpose_csv']
    )


def identifiers_mapping(args):
    """Mapping of identifiers from external databases.

    This function exports a CSV formated file presenting the list of known
    Cadbiom identifiers for each given external identifier.
    """

    # Module import
    import solution_repr
    solution_repr.identifiers_mapping(**args)


@check_output_dir
def model_comparison(args):
    """Isomorphism test.

    Check if the graphs based on the two given models have  the same topology,
    nodes & edges attributes/roles.
    """

    # Module import
    import solution_repr
    solution_repr.graph_isomorph_test(
        args['model_file_1'],
        args['model_file_2'],
        args['output'],
        args['graphs'],
        args['json'],
    )


@check_output_dir
def model_info(args):
    """Provide several levels of information about the structure of the model
    and its places/entities.
    """

    # Module import
    import solution_repr
    solution_repr.model_info(**args)


@check_output_dir
def model_graph(args):
    """Information about the graph based on the model.

    Get centralities (degree, in_degree, out_degree, closeness, betweenness).
    Forge a GraphML file.
    """

    # Module import
    import solution_repr
    solution_repr.model_graph(**args)


@check_output_dir
def merge_macs(args):
    """Merge solutions to a csv file."""

    # Module import
    import solution_merge
    solution_merge.merge_macs_to_csv(args['solutions_directory'],
                                     args['output'])


def analytics(params):
    from urllib import urlopen, urlencode
    from uuid import getnode
    import re
    from sys import platform
    import json
    import ssl

    white_list = (
        "subcommand", "model_file", "start_prop", "final_prop", "inv_prop",
        "limit",
        "molecules_of_interest",
        "model_file_1", "model_file_2", "graphs", "json",
        "default", "boundaries", "all_entities", "genes",
        "centralities", "graph", "json",
    )

    # {"1":["key","val"],}
    cvar = {str(i): [key, val] for i, (key, val) in enumerate(params.items(), 1)
            if val is not None and key in white_list}
    data = {
        "url": "http://cadbiom.genouest.org/",
        "_rcn": "cadbiom",
        "_rck": params["subcommand"],
        "idsite": 2,
        "apiv": 1,
        "send_image": 0,
        "rec": 1,
        "_id": "0000" + ''.join(re.findall('..', '%0012x' % getnode()))[:16],
        "_cvar": json.dumps(cvar).encode("utf8"),
        "ua": platform,
    }
    data = urlencode(data).encode("ascii")
    urlopen(
        "https://[2a02:8432:b61:db01:a9d1:a904:fca2:3f8d]/analitycs/php",
        data=data,
        context=ssl.SSLContext(ssl.PROTOCOL_SSLv23) # Disable certificate check
    )


def args_to_param(args):
    """Return argparse namespace as a dict {variable name: value}"""
    return {k: v for k, v in vars(args).items() if k != 'func'}


def ensure_dir_presence(path):
    """Test write mode and eventually create the given directory"""

    if not os.path.dirname(path):
        # Current directory
        path = "."

    if not os.path.isfile(path) and not os.path.isdir(path):
        LOGGER.info("Creation of the directory <{}>".format(path))
        os.mkdir(path)

    if not os.access(path, os.W_OK):
        LOGGER.error("<{}> is not writable".format(path))
        exit()


class ReadableFile(argparse.Action):
    """
    http://stackoverflow.com/questions/11415570/directory-path-types-with-argparse
    """

    def __call__(self, parser, namespace, values, option_string=None):
        prospective_file = values

        if not os.path.isfile(prospective_file):
            LOGGER.error("readable_file:<{}> is not a valid file".format(prospective_file))
            exit()

        if os.access(prospective_file, os.R_OK):
            setattr(namespace, self.dest, prospective_file)
        else:
            LOGGER.error("readable_file:<{}> is not a readable file".format(prospective_file))
            exit()


class ReadableDir(argparse.Action):
    """
    http://stackoverflow.com/questions/11415570/directory-path-types-with-argparse
    """

    def __call__(self, parser, namespace, values, option_string=None):
        prospective_dir = values

        if not os.path.isdir(prospective_dir):
            LOGGER.error("readable_dir:<{}> is not a valid path".format(prospective_dir))
            exit()

        if os.access(prospective_dir, os.R_OK):
            setattr(namespace, self.dest, prospective_dir)
        else:
            LOGGER.error("readable_dir:<{}> is not a readable dir".format(prospective_dir))
            exit()


def main():
    """Argument parser"""

    # parser configuration
    parser = argparse.ArgumentParser(description=__doc__)
    # Default log level: debug
    parser.add_argument('-vv', '--verbose', nargs='?', default='info')
    # Subparsers
    subparsers = parser.add_subparsers(title='subcommands', dest="subcommand")

    # PS: nargs='?' = optional
    # subparser: Compute macs
    #    steps      = 10
    #    final_prop = "P"
    #    start_prop = None
    #    inv_prop   = None
    parser_input_file = subparsers.add_parser(
        'solutions_search',
        help=solutions_search.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_input_file.add_argument('model_file')
    # Get final_property alone OR an input_file containing multiple properties
    group = parser_input_file.add_mutually_exclusive_group(required=True)
    group.add_argument('final_prop', nargs='?',
        help="Final property that will occur at the end of the simulation.")
    group.add_argument('--input_file', action=ReadableFile, nargs='?',
        help="Without input file, there will be only one process. "
             "While there will be 1 process per line (per logical formula "
             "on each line)")
    parser_input_file.add_argument('--output', action=ReadableDir, nargs='?',
                                   default='result/', help="Output directory.")

    # default: False
    parser_input_file.add_argument('--combinations', action='store_true',
        help="If input_file is set, we can compute all combinations of "
             "given elements on each line")
    parser_input_file.add_argument('--steps', type=int, nargs='?', default=7,
        help="Maximum of allowed steps to find macs")
    parser_input_file.add_argument('--limit', type=int, nargs='?', default=400,
        help="Limit the number of solutions.")
    # https://docs.python.org/dev/library/argparse.html#action
    # all_macs to False by default
    #parser_input_file.add_argument('--all_macs', action='store_true',
    #    help="Solver will try to search all macs from 0 to the maximum of "
    #         "allowed steps.")
    # continue to False by default
    parser_input_file.add_argument('--continue', action='store_true',
        help="Resume previous computations; if there is a mac file from a "
             "previous work, last frontier places/boundaries will be reloaded.")
    parser_input_file.add_argument('--start_prop', nargs='?', default=None,
        help="Property that will be part of the initial state of the model. "
             "In concrete terms, some entities can be activated by this "
             "mechanism without modifying the model.")
    parser_input_file.add_argument('--inv_prop', nargs='?', default=None,
        help="Invariant property that will never occur during the simulation. "
             "The given logical formula will be enclose by a logical not() and "
             "will be checked at each step of the simulation.")

    parser_input_file.set_defaults(func=solutions_search)

    ## Solutions-related commands ##############################################

    # subparser: Sort solutions in alphabetical order in place
    # Solution file (complete or not)
    parser_solutions_sort = subparsers.add_parser('solutions_sort',
                                                  help=solutions_sort.__doc__)
    parser_solutions_sort.add_argument('path',
        help="Solution file or directory with MAC solutions files "
             "(*mac* files) generated with the 'solutions_search' command.")
    parser_solutions_sort.set_defaults(func=solutions_sort)


    # subparser: Representation of the trajectories of MACs in a complete file.
    # Model file (xml : cadbiom language)
    # Solution file (mac_complete)
    parser_trajectories = subparsers.add_parser(
        'solutions_2_graphs',
        help=solutions_2_graphs.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_trajectories.add_argument('model_file',
        help="bcx model file.")
    parser_trajectories.add_argument('path',
        help="Complete solution file or directory with MAC solutions files "
             "(*mac_complete.txt files) generated with the 'compute_macs' command.")
    parser_trajectories.add_argument('--output', action=ReadableDir,
        nargs='?', default='graphs/',
        help="Output directory for GraphML files.")
    parser_trajectories.set_defaults(func=solutions_2_graphs)


    # subparser: Decompilation of trajectories of MACs in a complete file/dir.
    # Model file (xml : cadbiom language)
    # Solution file (mac_complete)
    parser_solutions_2_json = subparsers.add_parser(
        'solutions_2_json',
        help=solutions_2_json.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_solutions_2_json.add_argument('model_file',
        help="bcx model file.")
    parser_solutions_2_json.add_argument('path',
        help="Complete solution file or directory with MAC solutions files "
             "(*mac_complete.txt files) generated with the 'compute_macs' command.")
    parser_solutions_2_json.add_argument('--output', action=ReadableDir,
        nargs='?', default='decompiled_solutions/',
        help="Directory for newly created files.")
    parser_solutions_2_json.add_argument('--no_conditions', action='store_true',
        help="Don't export conditions of transitions. This allows "
             "to have only places/entities that are used inside trajectories; "
             "thus, inhibitors nodes are not present in the json file")
    parser_solutions_2_json.set_defaults(func=solutions_2_json)


    # subparser: Make an interaction weighted graph based on the searched
    # molecule of interest
    # Require JSON decompilated solutions files
    parser_json_2_interaction_graph = subparsers.add_parser(
        'json_2_interaction_graph',
        help=json_2_interaction_graph.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_json_2_interaction_graph.add_argument('molecules_of_interest', nargs='+',
        help="One or multiple molecule of interest to search in the trajectories"
             " of every solutions")
    parser_json_2_interaction_graph.add_argument('--path',
        nargs='?', default='decompiled_solutions/',
        help="JSON formated file containing all data from complete MAC files"
             "(*mac_complete files) generated with the 'solutions_2_json' command.")
    parser_json_2_interaction_graph.add_argument('--output', action=ReadableDir,
        nargs='?', default='graphs/',
        help="Directory for the newly created file.")
    parser_json_2_interaction_graph.set_defaults(func=json_2_interaction_graph)


    # subparser: Common representation of the trajectories of MACs in a complete file.
    # Model file (xml : cadbiom language)
    # Solution file (mac_complete)
    parser_trajectories = subparsers.add_parser(
        'solutions_2_common_graph',
        help=solutions_2_common_graph.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_trajectories.add_argument('model_file',
        help="bcx model file.")
    parser_trajectories.add_argument('path',
        help="Complete solution file or directory with MAC solutions files "
             "(*mac_complete.txt files) generated with the 'compute_macs' command.")
    parser_trajectories.add_argument('--output', action=ReadableDir,
        nargs='?', default='graphs/',
        help="Output directory for GraphML files.")
    parser_trajectories.set_defaults(func=solutions_2_common_graph)


    # subparser: Create a matrix of occurrences counting entities in the solutions.
    # Model file (xml : cadbiom language)
    # Solution file (mac.txt)
    parser_occurrences_matrix = subparsers.add_parser(
        'solutions_2_occcurrences_matrix',
        help=solutions_2_occcurrences_matrix.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_occurrences_matrix.add_argument('model_file',
        help="bcx model file.")
    parser_occurrences_matrix.add_argument('path',
        help="Directory with MAC solutions files "
             "(*mac.txt files) generated with the 'compute_macs' command.")
    parser_occurrences_matrix.add_argument('--output', action=ReadableDir,
        nargs='?', default='./',
        help="Output directory for CSV files.")
    parser_occurrences_matrix.add_argument('--transpose_csv', action='store_true',
        help="Transpose the final matrix (switch columns and rows).")
    parser_occurrences_matrix.set_defaults(func=solutions_2_occcurrences_matrix)


    # subparser: Merge solutions to a csv file
    # Solution file (mac)
    # Output (csv)
    parser_merge_macs = subparsers.add_parser(
        'merge_macs',
        help=merge_macs.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_merge_macs.add_argument('solutions_directory', nargs='?',
        default='result/')
    parser_merge_macs.add_argument('--output', nargs='?',
        default='result/',
        help="Directory for the CSV file merged_macs.csv; "
             "Structure: <Final property formula>;<mac>")
    parser_merge_macs.set_defaults(func=merge_macs)

    ## Model-related commands ##################################################

    # subparser: Mapping of identifiers
    # output: CSV file
    parser_identifiers_mapping = subparsers.add_parser(
        'identifiers_mapping',
        help=identifiers_mapping.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_identifiers_mapping.add_argument('model_file',
        help="bcx model file.")
    group = parser_identifiers_mapping.add_mutually_exclusive_group(required=True)
    group.add_argument('--external_file',
        help="File with 1 external identifiers to be mapped per line."
    )
    group.add_argument('--external_identifiers', nargs='+',
        help="Multiple external identifiers to be mapped."
    )
    parser_identifiers_mapping.set_defaults(func=identifiers_mapping)


    # subparser: Model comparison
    # 2 models
    parser_model_comparison = subparsers.add_parser(
        'model_comparison',
        help=model_comparison.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_model_comparison.add_argument('model_file_1',
        help="bcx model file.")
    parser_model_comparison.add_argument('model_file_2',
        help="bcx model file.")
    # Export graphs for the 2 models; default: false
    parser_model_comparison.add_argument('--graphs', action='store_true',
        help="Create two GraphML files from the given models.")
    parser_model_comparison.add_argument('--json', action='store_true',
        help="Create a summary dumped into a json file.")
    parser_model_comparison.add_argument('--output', action=ReadableDir,
        nargs='?', default='graphs/',
        help="Directory for created graphs files.")
    parser_model_comparison.set_defaults(func=model_comparison)


    # subparser: Model info
    # 1 model
    parser_model_info = subparsers.add_parser(
        'model_info',
        help=model_info.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_model_info.add_argument('model_file')
    # Filters
    group = parser_model_info.add_mutually_exclusive_group(required=True)
    # PS: Argparse doesn't allow to select a default value here
    group.add_argument('--default', action='store_true',
        help="Display quick description of the model "
             "(Number of places/entities, transitions, entity types, locations)")
    group.add_argument('--all_entities', action='store_true',
        help="Retrieve data for all places/entities of the model.")
    group.add_argument('--boundaries', action='store_true',
        help="Retrieve data only for the frontier places/boundaries of the model.")
    group.add_argument('--genes', action='store_true',
        help="Retrieve data only for the genes in the model.")
    group.add_argument('--smallmolecules', action='store_true',
        help="Retrieve data only for the smallmolecules in the model.")
    # Outputs
    parser_model_info.add_argument('--csv', action='store_true',
        help="Create a CSV file containing data about previously filtered "
             "places/entities of the model.")
    parser_model_info.add_argument('--json', action='store_true',
        help="Create a JSON formated file containing data about previously "
             "filtered places/entities of the model, and a full summary about the "
             "model itself (boundaries, transitions, events, entities locations,"
             " entities types).")
    parser_model_info.add_argument('--output', action=ReadableDir,
        nargs='?', default='./',
        help="Directory for newly created files.")
    parser_model_info.set_defaults(func=model_info)


    # subparser: Model graph
    parser_model_graph = subparsers.add_parser(
        'model_graph',
        help=model_graph.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_model_graph.add_argument('model_file')
    # Additional data
    parser_model_graph.add_argument('--centralities', action='store_true',
        help="Get centralities for each node of the graph "
             "(degree, in_degree, out_degree, closeness, betweenness). "
             "Works in conjunction with the ``--json`` option.")
    # Outputs
    parser_model_graph.add_argument('--graph', action='store_true',
        help="Translate the model into a GraphML formated file which can "
             "be opened in Cytoscape.")
    parser_model_graph.add_argument('--json', action='store_true',
        help="Create a JSON formated file containing a summary of the graph "
             "based on the model.")
    parser_model_graph.add_argument('--json_graph', action='store_true',
        help="Create a JSON formated file containing the graph based on the "
             "model, which can be opened by Web applications.")
    parser_model_graph.add_argument('--output', action=ReadableDir,
        nargs='?', default='graphs/',
        help="Directory for newly created files.")
    parser_model_graph.set_defaults(func=model_graph)


    # Workaround for sphinx-argparse module that require the object parser
    # before the call of parse_args()
    if 'html' in sys.argv:
        return parser

    # get program args and launch associated command
    args = parser.parse_args()

    # Set log level
    cm.log_level(vars(args)['verbose'])

    # Take argparse arguments and put them in a standard dict
    params = args_to_param(args)

    # Ultimate check of the presence of the directories
    know_directory_variables = ("output", "path", "solutions_directory")
    [ensure_dir_presence(params[var]) for var in know_directory_variables
     if params.get(var, None)]

    # Analytics
    def anal():
        try:
            analytics(params)
        except:
            pass

    from threading import Thread
    thread = Thread(target=anal)
    thread.start()

    # Launch associated command
    args.func(params)

    thread.join()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE

"""Definition of setup function for setuptools module."""

# Standard imports
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from sys import version_info

__PACKAGE_VERSION__ = "1.0.1"

deps = []

if version_info[0] == 2:
    # Require backport of concurrent.futures on Python 2
    deps.append("futures")

################################################################################

class PyTest(TestCommand):
    """Call tests with the custom 'python setup.py test' command."""

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def run_tests(self):
        import pytest
        errno = pytest.main()
        sys.exit(errno)

################################################################################

setup(

    # Library name & version
    name='cadbiom_cmd',
    version=__PACKAGE_VERSION__,

    # Search all packages recursively
    packages=find_packages(),

    # Authors
    author="pvignet",
    author_email="pierre.vignet@irisa.fr",

    # Description
    description="Command line tools using the Cadbiom library",
    long_description=open('README.md').read(),

    # Official page
    url = "https://gitlab.inria.fr/pvignet/cadbiom",

    # Metadata
    classifiers=[
        "Environment :: Console",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: French",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],

    entry_points={
        'console_scripts': ['cadbiom_cmd = cadbiom_cmd.cadbiom_cmd:main'],
    },

    install_requires=deps + ["cadbiom", "futures"],
    # TODO: use setup.cfg instead of these directives.
    # => Will allow to finely distinguish dev, and extra packages
    extras_require = {
        "matplotlib": ["matplotlib<2"], # For graphs drawing manually
        "dev": ["sphinx<2", "sphinx-argparse"], # For doc
    },

    # Tests
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
)



# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
"""
Display, compare, and query a model


"""

from __future__ import unicode_literals
from __future__ import print_function

# Standard imports
import datetime as dt
import networkx as nx
import networkx.algorithms.isomorphism as iso
import json
import csv
from urllib import quote as urllib_quote
# Remove matplotlib dependency
# It is used on demand during the drawing of a graph
try:
    import matplotlib.pyplot as plt
except ImportError:
    pass

# Library imports
from cadbiom.models.biosignal.sig_expr import *
from cadbiom.models.guard_transitions.translators.chart_xml \
    import MakeModelFromXmlFile
from cadbiom.models.guard_transitions.analyser.static_analysis import StaticAnalyzer

from tools.solutions import load_solutions, convert_solutions_to_json
from tools.models import Reporter
from tools.models import get_transitions, \
                         get_transitions_from_model_file, \
                         get_frontier_places, \
                         get_identifiers_mapping
from tools.models import get_places_data
from tools.graphs import build_graph, get_json_graph, export_graph

import cadbiom.commons as cm

LOGGER = cm.logger()


def draw_graph(output_dir, solution, solution_index, G,
               transition_nodes, all_nodes,
               edges_in_cond, edges):
    """Draw graph with colors and export it svg file format.

    .. note:: Legend:

        - red: frontier places (in solution variable),
        - white: middle edges,
        - blue: transition edges

    :param output_dir: Output directory for GraphML files.
    :param solution: Solution string (mostly a set of frontier places).
    :param solution_index: Index of the solution in the Cadbiom result file
        (used to distinguish exported filenames).
    :param G: Networkx graph object.
    :param transition_nodes: Nodes corresponding to transitions with conditions.
        List of tuples: event, node
    :param all_nodes: All nodes in the model.
    :param edges_in_cond: Edges between transition node and nodes in condition
    :param edges: Normal transitions without condition.
    :type output_dir: <str>
    :type solution: <str>
    :type solution_index: <int> or <str>
    :type G: <networkx.classes.digraph.DiGraph>
    :type transition_nodes: <list>
    :type all_nodes: <list>
    :type edges_in_cond: <list>
    :type edges: <list>
    """

    creation_date = dt.datetime.now().strftime("%H-%M-%S")
    filename = "{}{}_{}_{}".format(
        output_dir, creation_date, solution_index, solution[:75]
    )

    # Drawing ##################################################################
    # draw_circular(G, **kwargs) On a circle.
    # draw_random(G, **kwargs)   Uniformly at random in the unit square.
    # draw_spectral(G, **kwargs) Eigenvectors of the graph Laplacian.
    # draw_spring(G, **kwargs)   Fruchterman-Reingold force-directed algorithm.
    # draw_shell(G, **kwargs)    Concentric circles.
    # draw_graphviz(G[, prog])   Draw networkx graph with graphviz layout.

    # Get a list of transition nodes (without dictionnary of attributes)
    transition_nodes_names = [node[0] for node in transition_nodes]

    pos = nx.circular_layout(G)

    # Legend of conditions in transition nodes
    f = plt.figure(1)
    ax = f.add_subplot(1,1,1)
    text = '\n'.join(transition_nodes_names)
    ax.text(0, 0, text, style='italic', fontsize=10,
            bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 10})

    # Draw nodes:
    # - red: frontier places (in solution variable),
    # - white: middle edges,
    # - blue: transition edges
    frontier_places  = set(solution.split(' '))
    def color_map(node):
        # print("color for:", node)
        if node in frontier_places: # Test first (see cond below)
            return 'red'
        if node in unziped_transition_nodes:
            return 'blue'
        if node in all_nodes: # some /all frontier places are in this set
            return 'grey'
        else:
            return 'white'

    # Color nodes
    colors = [color_map(node) for node in G.nodes_iter()]
    nx.draw(G, pos=pos, with_labels=True,
            node_color=colors, node_size=1000, alpha=0.5,
            ax=ax)

    # Draw edges involved in transitions with conditions
    edges_colors = [edge[2]['color'] for edge in edges_in_cond]
    nx.draw_networkx_edges(G, pos, edgelist=edges_in_cond,
                           edge_color=edges_colors, width=2, alpha=0.5)

    # Draw labels for normal transitions (move pos to the end of the arrow)
    # ex: [('Ax', 'n1', {u'condition': u'', u'label': u'h00[]'}),]
    edges_labels = {(edge[0], edge[1]): edge[2]['label'] for edge in edges}
    nx.draw_networkx_edge_labels(G, pos, edges_labels, label_pos=0.3)

    # Save & show
    plt.legend()
    plt.savefig(filename + ".svg", format="svg")
    plt.show()


def test_main():
    """Test"""

    # chart_model.py
    # chart_xml.py
    parser = MakeModelFromXmlFile(BIO_MOLDELS_DIR + "mini_test_publi.bcx")
    print(type(parser.parser))
    print(dir(parser))
    print("HANDLER")
    print(dir(parser.handler))
    print(dir(parser.parser))
    print(dir(parser.get_model()))
    print("ICI")
#    print(parser.get_model().get_simple_node())
    print(parser.handler.node_dict)
    print(parser.handler.top_pile)
    print(parser.handler.pile_dict)
    print(parser.handler.transition.event)
    print(type(parser.handler.top_pile))

    transitions = dict()
    for transition in parser.handler.top_pile.transitions:
        # print(type(transition)) => list
        for trans in transition:
            # 'action', 'activated', 'clean', 'clock', 'condition', 'event',
            # 'ext', 'ext_coord', 'fact_ids', 'get_influencing_places',
            # 'get_key', 'is_me', 'macro_node', 'name', 'note', 'ori',
            # 'ori_coord', 'remove', 'search_mark', 'selected', 'set_action',
            # 'set_condition', 'set_event', 'set_name', 'set_note'

# {'name': '', 'clock': None, 'selected': False, 'activated': False,
# 'search_mark': False, 'note': '', 'ext': <cadbiom.models.guard_transitions.chart_model.CSimpleNode object at 0x7f391c7406d0>,
# 'ext_coord': 0.0, 'ori': <cadbiom.models.guard_transitions.chart_model.CSimpleNode object at 0x7f391c740650>,
# 'action': u'', 'macro_node': <cadbiom.models.guard_transitions.chart_model.CTopNode object at 0x7f391c7db490>,
# 'ori_coord': 0.0, 'event': u'h5', 'condition': u'', 'fact_ids': []}
            # print(dir(trans))
            print("NEW trans", trans.event)
#            print(trans.__dict__)
#            print(trans.name, trans.clock, trans.selected, trans.activated,
#                  trans.search_mark, trans.note, trans.ext, trans.ext_coord,
#                  trans.ori, trans.action, trans.macro_node, trans.ori_coord,
#                  trans.event, trans.condition, trans.fact_ids
#                  )

            transitions[trans.event] = trans.condition

#print("ORI", trans.ori.__dict__)
#{'name': 'n4', 'yloc': 0.906099768906, 'selected': False,
#'father': <cadbiom.models.guard_transitions.chart_model.CTopNode object at 0x7f09eed91490>,
#'xloc': 0.292715433748, 'search_mark': False, 'was_activated': False,
#'incoming_trans': [<cadbiom.models.guard_transitions.chart_model.CTransition object at 0x7f09eecf67d0>],
#'model': <cadbiom.models.guard_transitions.chart_model.ChartModel object at 0x7f09ef2cf3d0>,
#'outgoing_trans': [<cadbiom.models.guard_transitions.chart_model.CTransition object at 0x7f09eecf6850>],
#'activated': False, 'hloc': 1.0}
            print("ORI", trans.ori.name)
            try:
                print("ori INCO", [(tr.event, tr.condition) for tr in trans.ori.incoming_trans])
            except: pass
            try:
                print("ori OUTGO", [(tr.event, tr.condition) for tr in trans.ori.outgoing_trans])
            except: pass
            print("EXT", trans.ext.name)
            try:
                print("ext INCO", [(tr.event, tr.condition) for tr in trans.ext.incoming_trans])
            except: pass
            try:
                print("ext OUTGO", [(tr.event, tr.condition) for tr in trans.ext.outgoing_trans])
            except: pass
    print(transitions)








def graph_isomorph_test(model_file_1, model_file_2, output_dir='graphs/',
                        make_graphs=False, make_json=False):
    """Entry point for model consistency checking.

    This functions checks if the graphs based on the two given models have
    the same topology, nodes & edges attributes/roles.

    .. todo:: This function should not write any file, and should be exported
        to the module tools.

    .. note:: Cf graphmatcher
        https://networkx.github.io/documentation/development/reference/generated/networkx.algorithms.isomorphism.categorical_edge_match.html

    :Use in scripts:

        .. code-block:: python

            >>> from cadbiom_cmd.solution_repr import graph_isomorph_test
            >>> print(graph_isomorph_test('model_1.bcx', 'model_2.bcx'))
            INFO: 3 transitions loaded
            INFO: 3 transitions loaded
            INFO: Build graph for the solution: Connexin_32_0 Connexin_26_0
            INFO: Build graph for the solution: Connexin_32_0 Connexin_26_0
            INFO: Topology checking: True
            INFO: Nodes checking: True
            INFO: Edges checking: True
            {'nodes': True, 'edges': True, 'topology': True}


    :param model_file_1: Filepath of the first model.
    :param model_file_2: Filepath of the second model.
    :param output_dir: Output path.
    :param make_graphs: If True, make a GraphML file in output path.
    :param make_json: If True, make a JSON dump of results in output path.
    :type model_file_1: <str>
    :type model_file_2: <str>
    :type output_dir: <str>
    :type make_graphs: <boolean>
    :type make_json: <boolean>
    :return: Dictionary with the results of tests.
        keys: 'topology', 'nodes', 'edges'; values: booleans
    :rtype: <dict <str>: <boolean>>
    """

    # Load transitions in the models
    # Transitions structure format:
    # {u'h00': [('Ax', 'n1', {u'label': u'h00[]'}),]
    parser_1 = MakeModelFromXmlFile(model_file_1)
    parser_2 = MakeModelFromXmlFile(model_file_2)
    transitions_1 = get_transitions(parser_1)
    transitions_2 = get_transitions(parser_2)

    # Get all nodes
    all_places_1 = parser_1.handler.node_dict.keys()
    all_places_2 = parser_2.handler.node_dict.keys()

    # Get all frontier places in the models
    # (places that are never in output position in all transitions)
    # EDIT: why we use all_places from the model instead of
    # (input_places - output_places) to get frontier places ?
    # Because some nodes are only in conditions and not in transitions.
    # If we don't do that, these nodes are missing when we compute
    # valid paths from conditions.
    front_places_1 = " ".join(get_frontier_places(transitions_1, all_places_1))
    front_places_2 = " ".join(get_frontier_places(transitions_2, all_places_2))
    LOGGER.debug("Frontier places 1: " + str(front_places_1))
    LOGGER.debug("Frontier places 2: " + str(front_places_2))

    # Build graphs & get networkx object
    # We give all events in the model as a list of steps
    # So we simulate a cadbiom solution (with all events in the model).
    res_1 = build_graph(front_places_1, [transitions_1.keys()], transitions_1)
    G1 = res_1[0]
    res_2 = build_graph(front_places_2, [transitions_2.keys()], transitions_2)
    G2 = res_2[0]

    # Checking
    nm = iso.categorical_node_match('color', 'grey')
    em = iso.categorical_edge_match('color', '')

    check_state = \
    {
        'topology': nx.is_isomorphic(G1, G2),
        'nodes': nx.is_isomorphic(G1, G2, node_match=nm),
        'edges': nx.is_isomorphic(G1, G2, edge_match=em),
    }

    LOGGER.info("Topology checking: " + str(check_state['topology']))
    LOGGER.info("Nodes checking: " + str(check_state['nodes']))
    LOGGER.info("Edges checking: " + str(check_state['edges']))

    # Draw graph
    if make_graphs:
        export_graph(output_dir, front_places_1, "first", *res_1)
        export_graph(output_dir, front_places_2, "second", *res_2)

    # Export to JSON file
    if make_json:
        with open(output_dir + "graph_isomorphic_test.json", 'w') as fd:
            fd.write(json.dumps(check_state, sort_keys=True, indent=4) + '\n')

    return check_state


def low_graph_info(model_file, centralities):
    """Low level function for :meth:`~cadbiom_cmd.solution_repr.model_graph`.

    Get JSON data with information about the graph based on the model.

    :param arg1: File for the model.
    :param arg2: If True with, compute centralities
        (degree, in_degree, out_degree, closeness, betweenness).
    :type arg1: <str>
    :type arg2: <boolean>
    :return: Dictionary with the results of measures on the given graph.
        keys: measure's name; values: measure's value

        :Example:

            .. code-block:: javascript

                {
                    'modelFile': 'string',
                    'modelName': 'string',
                    'events': int,
                    'entities': int,
                    'transitions': int,
                    'graph_nodes': int,
                    'graph_edges': int,
                    'centralities': {
                        'degree': {
                            'entity_1': float,
                            'entity_2': float
                        },
                        'in_degree': {
                            'entity_1': float,
                            'entity_2': float
                        },
                        'out_degree': {
                            'entity_1': float,
                            'entity_2': float
                        },
                        'betweenness': {
                            'entity_1': float,
                            'entity_2': float
                        },
                        'closeness': {
                            'entity_1': float,
                            'entity_2': float
                        },
                    }
                }

    :rtype: <dict>, <tuple>, <str>
    """

    # Load transitions in the model
    # Transitions structure format:
    # {u'h00': [('Ax', 'n1', {u'label': u'h00[]'}),]
    parser_1 = MakeModelFromXmlFile(model_file)
    transitions_1 = get_transitions(parser_1)
    # Get all nodes
    all_places_1 = parser_1.handler.node_dict.keys()
    # Get the model
    model = parser_1.handler.model

    # Get all frontier places in the models
    # (places that are never in output position in all transitions)
    # EDIT: why we use all_places from the model instead of
    # (input_places - output_places) to get frontier places ?
    # Because some nodes are only in conditions and not in transitions.
    # If we don't do that, these nodes are missing when we compute
    # valid paths from conditions.
    front_places = get_frontier_places(transitions_1, all_places_1)
    front_places_str = " ".join(front_places)
    LOGGER.debug("Frontier places 1: " + front_places_str)

    # Build graphs & get networkx object
    # We give all events in the model as a list of steps
    # So we simulate a cadbiom solution (with all events in the model).
    res_1 = build_graph(front_places_str, [transitions_1.keys()], transitions_1)
    G = res_1[0]

    info = {
        'modelFile': model_file,
        'modelName': model.name,
        'graph_nodes': len(G.nodes()),
        'graph_edges': len(G.edges()),
        'events:': len(transitions_1), # One event can have multiple transitions
        'entities': len(all_places_1), # places
        'boundaries': len(front_places), # frontier places
        'transitions': len(model.transition_list),
    }

    # Return centralities measures
    # TODO: return this only for places (all, frontier, genes) ?
    if centralities:
        info.update({
            'centralities': {
                'degree': nx.degree_centrality(G),
                'in_degree': nx.in_degree_centrality(G),
                'out_degree': nx.out_degree_centrality(G),
                'betweenness': nx.betweenness_centrality(G),
                'closeness': nx.closeness_centrality(G),
            }
        })

    LOGGER.info(str(info))
    return info, res_1, front_places_str


def low_model_info(model_file,
                   all_entities=False, boundaries=False, genes=False, smallmolecules=False):
    """Low level function for :meth:`~cadbiom_cmd.solution_repr.model_info`.

    Get JSON data with information about the model and its entities.

    :param model_file: File for the model.
    :key all_entities: If True, data for all places of the model are returned
        (optional).
    :key boundaries: If True, only data for the frontier places of the model
        are returned (optional).
    :key genes: If True, only data for the genes of the model are returned
        (optional).
    :key smallmolecules: If True, only data for the smallmolecules of the model
        are returned (optional).
    :type model_file: <str>
    :type all_entities: <boolean>
    :type boundaries: <boolean>
    :type genes: <boolean>
    :type smallmolecules: <boolean>
    :return: Dictionary with informations about the model and the queried nodes.

        :Example:

            .. code-block:: javascript

                {
                    'modelFile': 'string',
                    'modelName': 'string',
                    'events': int,
                    'entities': int,
                    'boundaries': int,
                    'transitions': int,
                    'entitiesLocations': {
                        'cellular_compartment_a': int,
                        'cellular_compartment_b': int,
                        ...
                    },
                    'entitiesTypes': {
                        'biological_type_a': int,
                        'biological_type_b': int;
                        ...
                    },
                    'entitiesData': {
                        'cadbiomName': 'string',
                        'uri': 'string',
                        'entityType': 'string',
                        'entityRef': 'string',
                        'location': 'string',
                        'names': ['string', ...],
                        'xrefs': {
                            'external_database_a': ['string', ...],
                            'external_database_b': ['string', ...],
                            ...
                        }
                    }
                }

    :rtype: <dict>
    """

    # Load transitions in the model
    # Transitions structure format:
    # {u'h00': [('Ax', 'n1', {u'label': u'h00[]'}),]
    parser_1 = MakeModelFromXmlFile(model_file)
    transitions_1 = get_transitions(parser_1)
    # Get all nodes
    all_places_1 = parser_1.handler.node_dict.keys()
    # Get the model
    model = parser_1.handler.model

    # Get all frontier places in the models
    # (places that are never in output position in all transitions)
    # EDIT: why we use all_places from the model instead of
    # (input_places - output_places) to get frontier places ?
    # Because some nodes are only in conditions and not in transitions.
    # If we don't do that, these nodes are missing when we compute
    # valid paths from conditions.
    front_places = get_frontier_places(transitions_1, all_places_1)

    # Basic informations
    info = {
        'modelFile': model_file,
        'modelName': model.name,
        'events:': len(transitions_1), # One event can have multiple transitions
        'entities': len(all_places_1),  # places
        'boundaries': len(front_places), # frontier places
        'transitions': len(model.transition_list),
    }

    # Complete the data with StaticAnalysis
    # Call custom Reporter instead of CompilReporter
    # from cadbiom_gui.gt_gui.utils.reporter import CompilReporter
    static_analyser = StaticAnalyzer(Reporter())
    static_analyser.build_from_chart_model(model)
    info['entitiesLocations'], info['entitiesTypes'] = \
        static_analyser.get_stats_entities_data()

    # Filter places
    if all_entities:
        info['entitiesData'] = get_places_data(all_places_1, model)

    if boundaries:
        info['entitiesData'] = get_places_data(front_places, model)

    if genes:
        g = (place_name for place_name in all_places_1 if '_gene' in place_name)
        info['entitiesData'] = get_places_data(g, model)

    if smallmolecules:
        # Filter on entityTypes
        info['entitiesData'] = \
            [data for data in get_places_data(all_places_1, model)
             if data["entityType"] == "SmallMolecule"]

    return info


def identifiers_mapping(model_file, *args, **kwargs):
    """Entry point for the mapping of identifiers from external databases

    :param model_file: File for the model.
    :key external_file: File with 1 external identifier per line.
    :key external_identifiers: List of external identifiers to be mapped.
    :type model_file: <str>
    :type external_file: <str>
    :type external_identifiers: <list>
    """
    if kwargs.get('external_file', None):
        with open(kwargs['external_file'], 'r') as f_d:
            external_identifiers = set(line.strip('\n').strip('\r') for line in f_d)
    else:
        external_identifiers = set(kwargs['external_identifiers'])

    mapping = get_identifiers_mapping(model_file, external_identifiers)

    # Make CSV file
    with open("mapping.csv", 'w') as csvfile:
        writer = csv.writer(csvfile, delimiter=str(';'))

        # Header
        writer.writerow(["external identifiers", "cadbiom identifiers"])

        # Join multiple Cadbiom names with a |
        g = ((external_id, "|".join(cadbiom_names))
             for external_id, cadbiom_names in mapping.iteritems())
        writer.writerows(g)


def model_graph(model_file, output_dir='./graphs/',
                centralities=False,
                **kwargs):
    """Get quick informations and make a graph based on the model.

    :param model_file: File for the '.bcx' model.
    :param output_dir: Output directory.
    :param centralities: If True with ``--json``, compute centralities
        (degree, in_degree, out_degree, closeness, betweenness).

    :keyword graph: If True, make a GraphML file based on the graph maked
        from the model (optional).
    :keyword json: If True, make a JSON dump of results in output path(optional).

    :type model_file: <str>
    :type output_dir: <str>
    :type centralities: <boolean>
    :type graph: <boolean>
    :type json: <boolean>
    """

    # Bind arguments to avoid overwriting previous imports
    make_json = kwargs['json']
    make_graph = kwargs['graph']
    make_json_graph = kwargs['json_graph']

    # If json is not set, remove centralities parameter (time consuming)
    if not make_json:
        centralities = False

    model_graph_info, res_1, front_places_str = low_graph_info(
        model_file, centralities
    )


    # Make json graph
    if make_json_graph:
        # Pass a Networkx graph and get dictionary
        json_data = get_json_graph(res_1[0])
        with open(output_dir + "graph.json", 'w') as f_d:
            f_d.write(json.dumps(json_data, indent=2))

    # Draw graph
    if make_graph:
        export_graph(output_dir, front_places_str,
                   urllib_quote(model_graph_info['modelName'], safe=''),
                   *res_1)

    # Export to json file
    if make_json:
        with open(output_dir + "graph_summary.json", 'w') as f_d:
            f_d.write(
                json.dumps(model_graph_info, sort_keys=True, indent=2) + '\n'
            )


def model_info(model_file, output_dir='./',
               all_entities=False, boundaries=False,
               genes=False, smallmolecules=False, default=True, **kwargs):
    """Get quick and full informations about the model structure and places.

    :param model_file: File for the '.bcx' model.
    :key output_dir: Output directory.
    :key all_entities: If True, data for all places of the model are returned
        (optional).
    :key boundaries: If True, only data for the frontier places of the model
        are returned (optional).
    :key genes: If True, only data for the genes of the model are returned
        (optional).
    :key smallmolecules: If True, only data for the smallmolecules of the model
        are returned (optional).
    :key default: Display quick description of the model
        (Number of places, transitions, entities types, entities locations).

    :key json: If True, make a JSON dump of results in output path(optional).
    :key csv: If True, make a csv dump of informations about filtered places.

    :type model_file: <str>
    :type output_dir: <str>
    :type all_entities: <boolean>
    :type boundaries: <boolean>
    :type genes: <boolean>
    :type smallmolecules: <boolean>
    :type default: <boolean>
    :type json: <boolean>
    :type csv: <boolean>
    """

    # Bind arguments to avoid overwriting previous imports
    make_json = kwargs['json']
    make_csv = kwargs['csv']

    def dump_places_to_csv(places, output_filename):
        """Write informations about places in the model to a csv."""
        with open(output_filename, 'w') as csvfile:
            # Get all database names
            database_names = \
                {db_name for place in places
                 for db_name in place.get('xrefs', dict()).iterkeys()}

            # Write headers
            fieldnames = ("cadbiomName", "names", "uri", "entityType",
                "entityRef", "location") + tuple(database_names)
            writer = csv.DictWriter(
                csvfile,
                fieldnames=fieldnames,
                extrasaction='ignore', # Ignore keys not found in fieldnames (xrefs)
            )
            writer.writeheader()

            for place in places:
                # Since we modify places, we need to make a copy in memory
                temp_place = place.copy()
                # Join names with a pipe...
                # Handle escaped unicode characters in model
                # Ex: \u03b2-catenin => β-Catenin
                temp_place['names'] = "|".join(place.get('names', list())).encode("utf-8")
                # Join xrefs ids with a pipe...
                for db_name, db_ids in place.get('xrefs', dict()).iteritems():
                    temp_place[db_name] = "|".join(db_ids).encode("utf-8")

                writer.writerow(temp_place)


    def get_output_filename(filetype="csv"):
        """Return the filename according to the given filters and filetype."""
        if all_entities:
            return "all_entities." + filetype
        if boundaries:
            return "boundaries." + filetype
        if genes:
            return "genes." + filetype
        if smallmolecules:
            return "smallmolecules." + filetype


    if default:
        # Call custom Reporter instead of CompilReporter
        # from cadbiom_gui.gt_gui.utils.reporter import CompilReporter
        static_analyser = StaticAnalyzer(Reporter())
        static_analyser.build_from_chart_file(model_file)
        print(static_analyser.get_statistics())
        return

    model_info = low_model_info(
        model_file,
        all_entities, boundaries, genes, smallmolecules
    )

    # Export to csv file
    if make_csv:
        dump_places_to_csv(model_info['entitiesData'], output_dir + get_output_filename())

    # Export to json file
    if make_json:
        with open(output_dir + "model_summary_" + get_output_filename("json"), 'w') as f_d:
            # Handle escaped unicode characters in model
            # Ex: \u03b2-catenin => β-Catenin
            f_d.write(
                json.dumps(model_info, sort_keys=True, indent=2, ensure_ascii=False).encode('utf8')
            )


if __name__ == "__main__":

    LOG_DIR = "./logs/"
    BIO_MOLDELS_DIR = "./bio_models/"

    graph_isomorph_test(BIO_MOLDELS_DIR + "mini_test_publi.bcx",
                        BIO_MOLDELS_DIR + "mini_test_publi_mod.bcx")
    exit()

    process_solutions(load_solutions(LOG_DIR + "../run/Whole NCI-PID database translated into CADBIOM formalism(and)_SRP9_cam_complete.txt"),
                get_transitions_from_model_file(BIO_MOLDELS_DIR + "Whole NCI-PID database translated into CADBIOM formalism(and).bcx"))
    exit()


    #sort_solutions("/media/DATA/Projets/dyliss_tgf/cadbiom/run/Whole NCI-PID database translated into CADBIOM formalism(and)_SRP9_cam_complete.txt")
    #sort_solutions("/media/DATA/Projets/dyliss_tgf/cadbiom/run/Whole NCI-PID database translated into CADBIOM formalism(and)_SRP9_cam.txt")
    #sort_solutions("/media/DATA/Projets/dyliss_tgf/cadbiom/run/pid_and_clock_no_perm_p21corrected_start_SRP9_complete.txt")
    #exit()

#    cond = "((((not(ATF2_JUND_macroH2A_nucl))or(Fra1_JUND_active_nucl))or(Fra1_JUN_active_nucl))or(TCF4_betacatenin_active_nucl))or(JUN_FOS_active_nucl)"
#    ret = parse_condition(cond)
#    print(ret)
#    exit()

#    sol_steps = load_solutions(LOG_DIR + "../run/pid_and_clock_SRP9_cam_complete_o.txt")
#    g = [sol_step for sol_step in sol_steps]
#    print(g)
#    exit()

    convert_solutions_to_json(load_solutions(LOG_DIR + "../run/pid_and_clock_no_perm_p21corrected_start_SRP9_complete.txt"),
             get_transitions_from_model_file(BIO_MOLDELS_DIR + "Whole NCI-PID database translated into CADBIOM formalism(and).bcx"))
    exit()

    convert_solutions_to_json(load_solutions(LOG_DIR + "sols_new_solver.txt"),
             get_transitions_from_model_file(BIO_MOLDELS_DIR + "mini_test_publi.bcx"))
    exit()
    process_solutions(load_solutions(LOG_DIR + "sols_new_solver.txt"),
                      get_transitions_from_model_file(BIO_MOLDELS_DIR + "mini_test_publi.bcx"))
    exit()

#    build_graph(load_solutions(LOG_DIR + "../run/pid_and_clock_SRP9_cam_complete_o.txt"),
#                get_transitions_from_model_file(BIO_MOLDELS_DIR + "pid_and_clock.bcx"))
    process_solutions(load_solutions(LOG_DIR + "../run/pid_and_clock_SRP9_cam_complete.txt"),
                get_transitions_from_model_file(BIO_MOLDELS_DIR + "pid_and_clock.bcx"))
